/*
 * $Id: MainLoop.as,v 1.2 2006/07/17 05:06:54 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.dngn {
  import flash.display.Sprite;
  import flash.display.Bitmap;
  import flash.display.BitmapData;
  import flash.geom.Rectangle;
  import flash.events.Event;
  import flash.media.Sound;
  import flash.media.SoundChannel;
  import abagames.util.MouseState;
  import abagames.util.KeyState;
  import abagames.util.bulletml.BulletML;
  import abagames.util.bulletml.BulletPool;
  import abagames.util.bulletml.Bullet;
  import Dngn;

  /**
   * Main loop that updates a state.
   */
  public class MainLoop {
    private var bulletPool:BulletPoolImpl;
    private var prevTick:uint = 0;
    private var rootSprite:Sprite;
    private var buffer:BitmapData;
    private var screen:Bitmap;
    private var mouseState:MouseState;
    private var keyState:KeyState;
    private var cnt:Number;
    private var turretGroup:TurretGroup;
    private var barrages:Array;
    private var dots:Dots;
    private var turretCatalog:TurretCatalog;
    private var sound:Sound;
    private var channel:SoundChannel;
    private var nextChannelTick:Number;
    private var bgCnt:int;

    public function init
    (rootSprite:Sprite, buffer:BitmapData, screen:Bitmap, mouseState:MouseState, keyState:KeyState, sound:Sound):void {
      this.rootSprite = rootSprite;
      this.buffer = buffer;
      this.screen = screen;
      this.mouseState = mouseState;
      this.keyState = keyState;
      this.sound = sound;
      
      barrages = BarrageCatalog.createBarrages();
      dots = new Dots;
      bulletPool = new BulletPoolImpl;
      turretGroup = new TurretGroup(bulletPool);
      turretGroup.setBarrage(barrages[0]);
      var bx:int = 6;
      for each (var b:Barrage in barrages) {
	var bb:BarrageButton = new BarrageButton(b, turretGroup);
	bb.x = bx;
	bb.y = Dngn.screenHeight - BarrageButton.buttonHeight - 6;
	bb.width = BarrageButton.buttonWidth;
	bb.height = BarrageButton.buttonHeight;
	rootSprite.addChild(bb);
	bx += BarrageButton.buttonWidth + 6;
      }
      turretCatalog = new TurretCatalog;
      bx = 6;
      for each (var ts:TurretSpec in turretCatalog.specs) {
	var tb:TurretButton = new TurretButton(ts, turretGroup);
	tb.x = bx;
	tb.y = 6;
	tb.width = TurretButton.buttonWidth;
	tb.height = TurretButton.buttonHeight;
	rootSprite.addChild(tb);
	bx += TurretButton.buttonWidth + 6;
      }
      turretGroup.setSpec(turretCatalog.specs[0]);
      
      BulletImpl.screenWidth = Dngn.screenWidth;
      BulletImpl.screenHeight = Dngn.screenHeight
      BulletImpl.scale = 1;
      BulletImpl.rank = 1;
      BulletImpl.color = 0xffffff;
      BulletImpl.rect = new Rectangle(0, 0, 6, 6);
      TurretGroup.centerX = Dngn.screenWidth / 2;
      TurretGroup.centerY = Dngn.screenHeight / 3;
      Turret.color = 0xff8844;
      Turret.rect = new Rectangle(0, 0, 5, 5);
      cnt = 0;
      
      channel = sound.play(0, 99999);
      nextChannelTick = 0;
      bgCnt = 0;
    }

    public function update():void {
      cnt++;
      buffer.fillRect(buffer.rect, 0);
      var hs:Boolean = false;
      if (channel.position < nextChannelTick - 9999)
        nextChannelTick = 0;
      if (channel.position > nextChannelTick) {
	hs = true;
        if (bgCnt % 4== 0)
	  dots.setBackground();
        bgCnt++;
        nextChannelTick += 115.4;
      }
      BulletImpl.updateDotColor();
      dots.update(buffer);
      turretGroup.setMousePos(mouseState.x, mouseState.y);
      if (hs)
	turretGroup.update(buffer, dots);
      else
	turretGroup.update(buffer);
      if (mouseState.isClicked() &&
	  mouseState.y > TurretButton.buttonHeight + 6 &&
	  mouseState.y < Dngn.screenHeight - BarrageButton.buttonHeight - 6)
        turretGroup.mouseClicked();
      turretGroup.wheelMoved(mouseState.readDelta());
      if (keyState.right)
	turretGroup.wheelMoved(1);
      if (keyState.left)
	turretGroup.wheelMoved(-1);
      if (hs)
	bulletPool.run(buffer, dots);
      else
	bulletPool.run(buffer);
    }
  }
}
