/*
 * $Id: TurretCatalog.as,v 1.1.1.1 2006/07/17 03:27:32 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.dngn {
  /**
   * Catalog for TurretSpec.
   */
  public class TurretCatalog {
    public var specs:Array;

    public function TurretCatalog() {
      specs = new Array;
      specs.push(new RollTurrets);
      specs.push(new LineTurrets);
      specs.push(new EllipseRollTurrets);
      specs.push(new TwoWayRollTurrets);
      specs.push(new RopeTurrets);
      specs.push(new RingsRollTurrets);
      specs.push(new SquareTurrets);
      specs.push(new NarrowRollTurrets);
    }
  }
}

import abagames.dngn.Turret;
import abagames.dngn.TurretGroup;
import abagames.dngn.TurretSpec;
import abagames.dngn.BulletPoolImpl;

class RollTurrets implements TurretSpec {
  public function init(bullets:BulletPoolImpl):Array {
    var turrets:Array = new Array;
    for (var i:int = 0; i < 4; i++) {
      var t:Turret = new Turret(bullets);
      turrets.push(t);
      t.setSpec(this);
      t.deg = i * Math.PI / 2;
    }
    return turrets;
  }

  public function setMousePos(turrets:Array, x:int, y:int):void {
    var ox:Number = x - TurretGroup.centerX;
    var oy:Number = y - TurretGroup.centerY;
    var d:Number = Math.atan2(ox, oy);
    var r:Number = Math.sqrt(ox * ox + oy * oy);
    for each (var t:Turret in turrets) {
      t.trad = r;
      t.degVel = -d * 0.05;
    }
  }

  public function updateTurret(t:Turret):void {
    t.rad += (t.trad - t.rad) * 0.05;
    t.degVel += (t.tdegVel - t.degVel) * 0.07;
    t.deg += t.degVel;
    t.x = TurretGroup.centerX + Math.sin(t.deg) * t.rad;
    t.y = TurretGroup.centerY - Math.cos(t.deg) * t.rad;
  }
}

class LineTurrets implements TurretSpec {
  public function init(bullets:BulletPoolImpl):Array {
    var turrets:Array = new Array;
    for (var i:int = 0; i < 6; i++) {
      var t:Turret = new Turret(bullets);
      turrets.push(t);
      t.setSpec(this);
      if (i < 3) {
	t.posr = i * 0.33 + 0.34;
      } else {
	t.posr = (i - 3) * -0.33 - 0.34;
	t.fdv = -1;
      }
    }
    return turrets;
  }

  public function setMousePos(turrets:Array, x:int, y:int):void {
    var lx:Number = x;
    if (lx > TurretGroup.centerX)
      lx = TurretGroup.centerX - (lx - TurretGroup.centerX);
    var ox:Number = lx - TurretGroup.centerX;
    var oy:Number = y - TurretGroup.centerY;
    for each (var t:Turret in turrets) {
      t.tx = ox * t.posr + TurretGroup.centerX;
      if (t.posr < 0) {
	t.ty = -oy * t.posr + TurretGroup.centerY;
	t.deg = Math.atan2(t.x - TurretGroup.centerX, - t.y + TurretGroup.centerY) + Math.PI / 2;
      } else {
	t.ty = oy * t.posr + TurretGroup.centerY;
	t.deg = Math.atan2(t.x - TurretGroup.centerX, -t.y + TurretGroup.centerY) - Math.PI / 2;
      }
    }
  }

  public function updateTurret(t:Turret):void {
    t.x += (t.tx - t.x) * 0.1;
    t.y += (t.ty - t.y) * 0.1;
  }
}

class EllipseRollTurrets implements TurretSpec {
  public function init(bullets:BulletPoolImpl):Array {
    var turrets:Array = new Array;
    for (var i:int = 0; i < 5; i++) {
      var t:Turret = new Turret(bullets);
      turrets.push(t);
      t.setSpec(this);
      t.deg = i * Math.PI * 2 / 5;
    }
    return turrets;
  }

  public function setMousePos(turrets:Array, x:int, y:int):void {
    var ox:Number = x - TurretGroup.centerX;
    var oy:Number = y - TurretGroup.centerY;
    var d:Number = Math.atan2(-ox, oy);
    var r:Number = Math.sqrt(ox * ox + oy * oy);
    for each (var t:Turret in turrets) {
      t.trad = r;
      t.ttrDeg = d;
    }
  }

  public function updateTurret(t:Turret):void {
    t.rad += (t.trad - t.rad) * 0.05;
    t.trDeg += (t.ttrDeg - t.trDeg) * 0.1;
    t.deg += 0.02;//6;
    var tox:Number = Math.sin(t.deg) * t.rad * 0.33;
    var toy:Number = -Math.cos(t.deg) * t.rad;
    var ox:Number = tox * Math.cos(t.trDeg) - toy * Math.sin(t.trDeg);
    var oy:Number = tox * Math.sin(t.trDeg) + toy * Math.cos(t.trDeg);
    t.x = TurretGroup.centerX + ox;
    t.y = TurretGroup.centerY + oy;
  }
}

class TwoWayRollTurrets implements TurretSpec {
  public function init(bullets:BulletPoolImpl):Array {
    var turrets:Array = new Array;
    for (var i:int = 0; i < 6; i++) {
      var t:Turret = new Turret(bullets);
      turrets.push(t);
      t.setSpec(this);
      t.deg = i * Math.PI * 2 / 3;
      if (i > 2)
	t.fdv = -1;
    }
    return turrets;
  }

  public function setMousePos(turrets:Array, x:int, y:int):void {
    var ox:Number = x - TurretGroup.centerX;
    var oy:Number = y - TurretGroup.centerY;
    var d:Number = Math.atan2(ox, oy);
    var r:Number = Math.sqrt(ox * ox + oy * oy);
    for each (var t:Turret in turrets) {
      t.trad = r;
      t.degVel = -d * 0.05 * t.fdv;
    }
  }

  public function updateTurret(t:Turret):void {
    t.rad += (t.trad - t.rad) * 0.05;
    t.degVel += (t.tdegVel - t.degVel) * 0.07;
    t.deg += t.degVel;
    t.x = TurretGroup.centerX + Math.sin(t.deg) * t.rad;
    t.y = TurretGroup.centerY - Math.cos(t.deg) * t.rad;
  }
}

class RopeTurrets implements TurretSpec {
  public function init(bullets:BulletPoolImpl):Array {
    var turrets:Array = new Array;
    for (var i:int = 0; i < 6; i++) {
      var t:Turret = new Turret(bullets);
      turrets.push(t);
      t.setSpec(this);
      if (i < 3) {
	t.posr = i * 0.3 + 0.15;
      } else {
	t.posr = (i - 3) * -0.3 - 0.15;
	t.fdv = -1;
      }
    }
    return turrets;
  }

  public function setMousePos(turrets:Array, x:int, y:int):void {
    var ox:Number = x - TurretGroup.centerX;
    var oy:Number = y - TurretGroup.centerY;
    var d:Number = Math.atan2(-ox, oy);
    var r:Number = Math.sqrt(ox * ox + oy * oy);
    for each (var t:Turret in turrets) {
      t.trad = r;
      t.ttrDeg = d;
    }
  }

  public function updateTurret(t:Turret):void {
    t.rad += (t.trad - t.rad) * 0.05;
    t.trDeg += (t.ttrDeg - t.trDeg) * 0.1;
    t.cdeg += 0.03;
    var tox:Number = Math.sin(t.cdeg) * t.rad * 0.33;
    var toy:Number = -Math.cos(t.cdeg) * t.rad;
    var ox:Number = tox * Math.cos(t.trDeg) - toy * Math.sin(t.trDeg);
    var oy:Number = tox * Math.sin(t.trDeg) + toy * Math.cos(t.trDeg);
    var cx:Number = TurretGroup.centerX + ox;
    var cy:Number = TurretGroup.centerY + oy;
    var loy:Number = TurretGroup.centerY - cy;
    var lox:Number;
    if (t.posr < 0) {
      lox = TurretGroup.centerX * 0.4 - cx;
      t.tx = -lox * t.posr + cx;
      t.ty = -loy * t.posr + cy;
      t.deg = Math.atan2(TurretGroup.centerX - t.x, -(TurretGroup.centerY - t.y)) - t.trDeg;
    } else {
      lox = TurretGroup.centerX * 1.6 - cx;
      t.tx = lox * t.posr + cx;
      t.ty = loy * t.posr + cy;
      t.deg = Math.atan2(TurretGroup.centerX - t.x, -(TurretGroup.centerY - t.y)) - t.trDeg;
    }
    t.x += (t.tx - t.x) * 0.1;
    t.y += (t.ty - t.y) * 0.1;
  }
}

class RingsRollTurrets implements TurretSpec {
  public function init(bullets:BulletPoolImpl):Array {
    var turrets:Array = new Array;
    for (var i:int = 0; i < 4; i++) {
      var t:Turret = new Turret(bullets);
      turrets.push(t);
      t.setSpec(this);
      t.deg = i * Math.PI;
    }
    return turrets;
  }

  public function setMousePos(turrets:Array, x:int, y:int):void {
    var ox:Number = x - TurretGroup.centerX;
    var oy:Number = y - TurretGroup.centerY;
    var d:Number = Math.atan2(ox, oy);
    var r:Number = Math.sqrt(ox * ox + oy * oy);
    var i:int = 0;
    for each (var t:Turret in turrets) {
      if (i < 2) {
	t.trad = r;
	t.degVel = -d * 0.05;
      } else {
	t.trad = r * 0.5;
	t.degVel = d * 0.05;
      }
      i++;
    }
  }

  public function updateTurret(t:Turret):void {
    t.rad += (t.trad - t.rad) * 0.05;
    t.degVel += (t.tdegVel - t.degVel) * 0.07;
    t.deg += t.degVel;
    t.x = TurretGroup.centerX + Math.sin(t.deg) * t.rad;
    t.y = TurretGroup.centerY - Math.cos(t.deg) * t.rad;
  }
}

class SquareTurrets implements TurretSpec {
  public function init(bullets:BulletPoolImpl):Array {
    var turrets:Array = new Array;
    for (var i:int = 0; i < 4; i++) {
      var t:Turret = new Turret(bullets);
      turrets.push(t);
      t.setSpec(this);
      switch (i) {
      case 0:
	t.posr = 1;
	break;
      case 1:
	t.posr = -1;
	t.fdv = -1;
	break;
      case 2:
	t.posr = 0.5;
	break;
      case 3:
	t.posr = -0.5;
	t.fdv = -1;
	break;
      }
      t.deg = Math.PI;
    }
    return turrets;
  }

  public function setMousePos(turrets:Array, x:int, y:int):void {
    var lx:Number = x;
    if (lx > TurretGroup.centerX)
      lx = TurretGroup.centerX - (lx - TurretGroup.centerX);
    var ox:Number = lx - TurretGroup.centerX;
    var oy:Number = y - TurretGroup.centerY;
    var i:int = 0;
    for each (var t:Turret in turrets) {
      t.cdeg += 0.03;
      t.tx = ox * t.posr * Math.cos(t.cdeg) + TurretGroup.centerX;
      if (i < 2)
	t.ty = oy + TurretGroup.centerY;
      else
	t.ty = -oy + TurretGroup.centerY;
      i++;
    }
  }

  public function updateTurret(t:Turret):void {
    t.x += (t.tx - t.x) * 0.1;
    t.y += (t.ty - t.y) * 0.1;
  }
}

class NarrowRollTurrets implements TurretSpec {
  public function init(bullets:BulletPoolImpl):Array {
    var turrets:Array = new Array;
    for (var i:int = 0; i < 5; i++) {
      var t:Turret = new Turret(bullets);
      turrets.push(t);
      t.setSpec(this);
      t.deg = i * Math.PI / 15;
    }
    return turrets;
  }

  public function setMousePos(turrets:Array, x:int, y:int):void {
    var ox:Number = x - TurretGroup.centerX;
    var oy:Number = y - TurretGroup.centerY;
    var d:Number = Math.atan2(ox, oy);
    var r:Number = Math.sqrt(ox * ox + oy * oy);
    for each (var t:Turret in turrets) {
      t.trad = r;
      t.degVel = -d * 0.05;
    }
  }

  public function updateTurret(t:Turret):void {
    t.rad += (t.trad - t.rad) * 0.05;
    t.degVel += (t.tdegVel - t.degVel) * 0.07;
    t.deg += t.degVel;
    t.x = TurretGroup.centerX + Math.sin(t.deg) * t.rad;
    t.y = TurretGroup.centerY - Math.cos(t.deg) * t.rad;
  }
}
