/*
 * $Id: Action.as,v 1.1.1.1 2006/07/17 03:27:32 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.util.bulletml {
  /**
   * Handling a BulletML action.
   */
  public class Action {
    public static const DEFAULT_SPEED:Number = 1;
    private var root:XML; // Root of a BulletML XML for finding a reference of Action/Bullet/FireRef.
    private var bml:XML;
    private var ip:int; // Instruction pointer.
    private var instructions:Array;  // Instructions (BulletML commands).
    private var parent:Action;
    private var child:Action; // Nested action.
    private var state:BulletState;
    private var waitCnt:int;
    private var repeatCnt:int;
    private var isVanished:Boolean;
    private var params:Array;

    public function init(bml:XML, root:XML, params:Array = null):void {
      parent = null;
      state = new BulletState;
      repeatCnt = 1;
      this.params = params;
      initFirst(bml, root);
    }

    public function initChild(bml:XML, root:XML, parent:Action, state:BulletState,
                              repeatCnt:int, params:Array):void {
      this.parent = parent;
      this.state = state;
      this.repeatCnt = repeatCnt;
      this.params = params;
      initFirst(bml, root);
    }

    private function initFirst(bml:XML, root:XML):void {
      this.root = root;
      this.bml = bml;
      instructions = new Array;
      for each (var i:XML in bml.*)
        instructions.push(i);
      ip = 0;
      waitCnt = 0;
      child = null;
      isVanished = false;
    }

    public function run(bullet:Bullet, bulletPool:BulletPool):void {
      if (isEnd())
        return;
      state.run(bullet);
      runInstructions(bullet, bulletPool);
    }

    private function runInstructions(bullet:Bullet, bulletPool:BulletPool):void {
      for (;;) {
        if (isInstructionsEnd())
          return;
        if (child != null) {
          child.runInstructions(bullet, bulletPool);
          if (child.isEnd())
            child = null;
          else
            return;
        }
        if (waitCnt > 0) {
          waitCnt--;
          return;
        }
        if (repeatCnt > 0) {
          var ci:XML = instructions[ip];
          var ciName:String = ci.name().toString();
          var a:XML;
          var tm:int;
          switch (ciName) {
          case "repeat":
            var rt:int = int(Expression.parse(ci.times.toString()).calc(params, bullet));
            if (ci.action != undefined) {
              child = new Action;
              child.initChild(ci.action[0], root, this, state, rt, params);
            } else if (ci.actionRef != undefined) {
              a = root.action.(@label == ci.actionRef[0].@label)[0];
              child = new Action;
              child.initChild(a, root, this, state, rt, parseParams(ci.actionRef[0], bullet, params));
            }
            break;
          case "action":
            child = new Action;
            child.initChild(ci, root, this, state, 1, params);
            break;
          case "actionRef":
            a = root.action.(@label == ci.@label)[0];
            child = new Action;
            child.initChild(a, root, this, state, 1, parseParams(ci, bullet, params));
            break;
          case "fireRef":
          case "fire":
            var rp:Array = params; // From here using rp for params.
            if (ciName == "fireRef") {
              var frp:Array = parseParams(ci, bullet, rp);
              if (frp != null)
                rp = frp;
              ci = root.fire.(@label == ci.@label)[0];
            }
            var deg:Number = getDeg(ci, bullet, state, rp);
            var sp:Number = getSpeed(ci, bullet, state, rp);
            var bl:XML;
            if (ci.bullet != undefined) {
              bl = ci.bullet[0];
            } else if (ci.bulletRef != undefined) {
              bl = root.bullet.(@label == ci.bulletRef.@label)[0];
              var brp:Array = parseParams(ci.bulletRef[0], bullet, rp);
              if (brp != null)
                rp = brp;
            }
            if (bl.direction != undefined)
              deg = getDeg(bl, bullet, state, rp);
            if (bl.speed != undefined)
              sp = getSpeed(bl, bullet, state, rp);
            state.setSequenceFireDeg(deg);
            state.setSequenceFireSpeed(sp);
            var bml:BulletML = new BulletML;
            bml.initPartial(bl, root, rp, bullet);
            bulletPool.addBullet(bml, bullet.getX(), bullet.getY(), deg, sp);
            break;
          case "changeSpeed":
            tm = int(Expression.parse(ci.term.toString()).calc(params, bullet));
            var spv:Number = getSpeedVelocity(ci, bullet, tm);
            state.changeSpeed(spv, tm);
            break;
          case "changeDirection":
            tm = int(Expression.parse(ci.term.toString()).calc(params, bullet));
            var dv:Number = getDirectionVelocity(ci, bullet, tm);
            state.changeDirection(dv, tm);
            break;
          case "accel":
            tm = int(Expression.parse(ci.term.toString()).calc(params, bullet));
            if (ci.horizontal != undefined) {
              var hv:Number = getHorizontalVelocity(ci, bullet, tm);
              state.accelHorizontal(hv, tm);
            }
            if (ci.vertical != undefined) {
              var vv:Number = getVerticalVelocity(ci, bullet, tm);
              state.accelVertical(vv, tm);
            }
            break;
          case "wait":
            waitCnt = int(Expression.parse(ci.toString()).calc(params, bullet));
            break;
          case "vanish":
            bullet.vanish();
            vanished();
            return;
          default:
            throw new Error("unknown command: " + ciName);
          }
          ip++;
          if (ip >= instructions.length) {
            repeatCnt--;
            ip = 0;
          }
        }
      }
    }

    public function isEnd():Boolean {
      return ((repeatCnt <= 0 && waitCnt <= 0 && child == null && state.isEnd()) || isVanished);
    }

    private function isInstructionsEnd():Boolean {
      return ((repeatCnt <= 0 && waitCnt <= 0 && child == null) || isVanished);
    }

    public function vanished():void {
      isVanished = true;
      if (parent != null)
        parent.vanished();
    }

    private function getDeg(ci:XML, bullet:Bullet, state:BulletState, params:Array):Number {
      var deg:Number;
      var d:Number = Expression.parse(ci.direction.toString()).calc(params, bullet);
      switch (ci.direction.@type.toString()) {
      case "absolute":
        deg = d;
        break;
      case "relative":
        deg = bullet.getDirection() + d;
        break;
      case "sequence":
        deg = state.getSequenceFireDeg() + d;
        break;
      case "aim":
      default:
        deg = bullet.getAimDirection();
        if (ci.direction != undefined)
          deg += d;
        break;
      }
      return deg;
    }

    private function getDirectionVelocity(ci:XML, bullet:Bullet, term:Number):Number {
      var dv:Number;
      var d:Number = Expression.parse(ci.direction.toString()).calc(params, bullet);
      switch (ci.direction.@type.toString()) {
      case "absolute":
        dv = BulletState.adjustDeg(d - bullet.getDirection()) / term;
        break;
      case "relative":
        dv = d / term;
        break;
      case "sequence":
        dv = d;
        break;
      case "aim":
      default:
        dv = BulletState.adjustDeg(d + bullet.getAimDirection() - bullet.getDirection()) / term;
        break;
      }
      return dv;
    }

    private function getSpeed(ci:XML, bullet:Bullet, state:BulletState, params:Array):Number {
      var sp:Number;
      var s:Number = Expression.parse(ci.speed.toString()).calc(params, bullet);
      switch (ci.speed.@type.toString()) {
      case "relative":
        sp = bullet.getSpeed() + s;
        break;
      case "sequence":
        sp = state.getSequenceFireSpeed() + s;
        break;
      case "absolute":
      default:
        if (ci.speed == undefined)
          sp = DEFAULT_SPEED;
        else
          sp = s;
        break;
      }
      return sp;
    }

    private function getSpeedVelocity(ci:XML, bullet:Bullet, term:Number):Number {
      var spv:Number;
      var s:Number = Expression.parse(ci.speed.toString()).calc(params, bullet);
      switch (ci.speed.@type.toString()) {
      case "relative":
        spv = s / term;
        break;
      case "sequence":
        spv = s;
        break;
      case "absolute":
      default:
        if (ci.speed == undefined)
          spv = DEFAULT_SPEED - bullet.getSpeed();
        else
          spv = s - bullet.getSpeed();
        spv /= term;
        break;
      }
      return spv;
    }

    private function getHorizontalVelocity(ci:XML, bullet:Bullet, term:Number):Number {
      var hv:Number;
      var h:Number = Expression.parse(ci.horizontal.toString()).calc(params, bullet);
      switch (ci.horizontal.@type.toString()) {
      case "relative":
        hv = h / term;
        break;
      case "sequence":
        hv = h;
        break;
      case "absolute":
      default:
        hv = h - bullet.getHorizontal();
        hv /= term;
        break;
      }
      return hv;
    }

    private function getVerticalVelocity(ci:XML, bullet:Bullet, term:Number):Number {
      var vv:Number;
      var v:Number = Expression.parse(ci.vertical.toString()).calc(params, bullet);
      switch (ci.vertical.@type.toString()) {
      case "relative":
        vv = v / term;
        break;
      case "sequence":
        vv = v;
        break;
      case "absolute":
      default:
        vv = v - bullet.getVertical();
        vv /= term;
        break;
      }
      return vv;
    }
    
    public function parseParams(ref:XML, bullet:Bullet, params:Array):Array {
      var rp:Array = new Array;
      for each (var v:XML in ref.param)
        rp.push(Expression.parse(v.toString()).calc(params, bullet));
      if (rp.length <= 0)
        rp = null;
      return rp;
    }
  }
}

import abagames.util.bulletml.Bullet;
import abagames.util.bulletml.Action;

/**
 * State of a bullet inherited by a child action.
 */
class BulletState {
  private var spVel:Number; // For changeSpeed.
  private var spVelCnt:int;
  private var degVel:Number; // For changeDirection.
  private var degVelCnt:int;
  private var acxVel:Number; // For accel.
  private var acxVelCnt:int;
  private var acyVel:Number;
  private var acyVelCnt:int;
  private var seqFireDeg:Number; // For a sequence type fire.
  private var seqFireSpeed:Number;

  public function BulletState() {
    spVelCnt = degVelCnt = acxVelCnt = acyVelCnt = 0;
    seqFireDeg = 0;
    seqFireSpeed = Action.DEFAULT_SPEED;
  }

  public function run(bullet:Bullet):void {
    if (spVelCnt > 0) {
      spVelCnt--;
      bullet.setSpeed(bullet.getSpeed() + spVel);
    }
    if (degVelCnt > 0) {
      degVelCnt--;
      bullet.setDirection(adjustDeg(bullet.getDirection() + degVel));
    }
    if (acxVelCnt > 0) {
      acxVelCnt--;
      bullet.setHorizontal(bullet.getHorizontal() + acxVel);
    }
    if (acyVelCnt > 0) {
      acyVelCnt--;
      bullet.setVertical(bullet.getVertical() + acyVel);
    }
  }

  public function isEnd():Boolean {
    return (spVelCnt <= 0 && degVelCnt <= 0 && acxVelCnt <= 0 && acyVelCnt <=0);
  }

  public function getSequenceFireDeg():Number {
    return seqFireDeg;
  }

  public function setSequenceFireDeg(deg:Number):void {
    seqFireDeg = deg;
  }

  public function getSequenceFireSpeed():Number {
    return seqFireSpeed;
  }

  public function setSequenceFireSpeed(sp:Number):void {
    seqFireSpeed = sp;
  }

  public function changeSpeed(spv:Number, tm:Number):void {
    spVelCnt = tm;
    spVel = spv;
  }

  public function changeDirection(dv:Number, tm:Number):void {
    degVelCnt = tm;
    degVel = dv;
  }

  public function accelHorizontal(hv:Number, tm:Number):void {
    acxVelCnt = tm;
    acxVel = hv;
  }

  public function accelVertical(vv:Number, tm:Number):void {
    acyVelCnt = tm;
    acyVel = vv;
  }

  public static function adjustDeg(d:Number):Number {
    d %= 360;
    if (d > 180)
      return d - 360;
    else if (d < -180)
      return d + 360;
    else
      return d;
  }
}
